import numpy as np
from matplotlib.patches import Circle, RegularPolygon
from matplotlib.path import Path
from matplotlib.projections import register_projection
from matplotlib.projections.polar import PolarAxes
from matplotlib.spines import Spine
from matplotlib.transforms import Affine2D


def spider_factory(num_vars, frame='circle'):
    """
    Taken from https://matplotlib.org/stable/gallery/specialty_plots/radar_chart.html
    Create a radar chart with `num_vars` axes.

    This function creates a RadarAxes projection and registers it.

    Parameters
    ----------
    num_vars : int
        Number of variables for radar chart.
    frame : {'circle', 'polygon'}
        Shape of frame surrounding axes.

    """
    # calculate evenly-spaced axis angles
    theta = np.linspace(0, 2 * np.pi, num_vars, endpoint=False)

    class RadarTransform(PolarAxes.PolarTransform):

        def transform_path_non_affine(self, path):
            # Paths with non-unit interpolation steps correspond to gridlines,
            # in which case we force interpolation (to defeat PolarTransform's
            # autoconversion to circular arcs).
            if path._interpolation_steps > 1:
                path = path.interpolated(num_vars)
            return Path(self.transform(path.vertices), path.codes)

    class RadarAxes(PolarAxes):

        name = 'radar'
        PolarTransform = RadarTransform

        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            # rotate plot such that the first axis is at the top
            self.set_theta_zero_location('N')

        def fill(self, *args, closed=True, **kwargs):
            """Override fill so that line is closed by default"""
            return super().fill(closed=closed, *args, **kwargs)

        def plot(self, *args, **kwargs):
            """Override plot so that line is closed by default"""
            lines = super().plot(*args, **kwargs)
            for line in lines:
                self._close_line(line)

        def _close_line(self, line):
            x, y = line.get_data()
            # FIXME: markers at x[0], y[0] get doubled-up
            if x[0] != x[-1]:
                x = np.append(x, x[0])
                y = np.append(y, y[0])
                line.set_data(x, y)

        def set_varlabels(self, labels):
            self.set_thetagrids(np.degrees(theta), labels)

        def _gen_axes_patch(self):
            # The Axes patch must be centered at (0.5, 0.5) and of radius 0.5
            # in axes coordinates.
            if frame == 'circle':
                return Circle((0.5, 0.5), 0.5)
            elif frame == 'polygon':
                return RegularPolygon((0.5, 0.5), num_vars,
                                      radius=.5, edgecolor="k")
            else:
                raise ValueError("Unknown value for 'frame': %s" % frame)

        def _gen_axes_spines(self):
            if frame == 'circle':
                return super()._gen_axes_spines()
            elif frame == 'polygon':
                # spine_type must be 'left'/'right'/'top'/'bottom'/'circle'.
                spine = Spine(axes=self,
                              spine_type='circle',
                              path=Path.unit_regular_polygon(num_vars))
                # unit_regular_polygon gives a polygon of radius 1 centered at
                # (0, 0) but we want a polygon of radius 0.5 centered at (0.5,
                # 0.5) in axes coordinates.
                spine.set_transform(Affine2D().scale(.5).translate(.5, .5)
                                    + self.transAxes)
                return {'polar': spine}
            else:
                raise ValueError("Unknown value for 'frame': %s" % frame)

    register_projection(RadarAxes)
    return theta


def get_polygon_corners(theta, bottom, top):
    """
    Retrieves the corners of a polygon encompassed by the shades borders to allow for proper shading

    :param theta: radius reference
    :param bottom: lower border of the shading. y values
    :param top: upper border of the shading. y values
    :return: x and y values of the polygon encompassed by the borders
    """
    x = []
    y = []
    for i in theta:
        x.append(i)
        x.append(i)
    for i in range(len(theta)):
        y.append(bottom[i])
        y.append(top[i])
    return x, y


def plot_spider_single_trace(ax, data, color, theta, zorder=0):
    """
    Plots a single trace on a spider plot (radar plot)

    :param ax: axis to plot on
    :param data: dataset to plot
    :param color: color to use to plot
    :param theta: radius reference
    :param zorder: z-order for stacking
    :return:
    """
    count_avg = data["mean"].values
    n_stat = data["lower_bound"].values
    n_stat[n_stat < 0] = 0
    p_stat = data["upper_bound"].values

    ax.plot(theta, count_avg, color=color, linewidth=3, zorder=zorder + 10)

    ax.scatter(theta, count_avg, color=color, s=50, zorder=zorder + 20)

    ax.fill_between(theta, n_stat, p_stat,
                    facecolor=color, alpha=0.4, zorder=zorder)

    ax.fill([theta[0], theta[0], theta[-1], theta[-1]], [p_stat[0], n_stat[0], n_stat[-1], p_stat[-1]],
            facecolor=color, alpha=0.4, zorder=zorder)



import yaml


# TESTME with config file
def read_config(path, converts_keys_to_string=False):
    """
    Reads structured config file defining a project.

    :param path: path to config
    :param converts_keys_to_string:
    :return: dict from a yaml file
    """

    try:
        with open(path, "r") as ymlfile:
            cfg = yaml.load(ymlfile, Loader=yaml.SafeLoader)
    except FileNotFoundError:
        raise FileNotFoundError(
            "Could not find the config file at " + path + " \n Please make sure the path is correct and the file exists")

    if converts_keys_to_string:
        keys2string(cfg)

    return cfg


def keys2string(d):
    """
    Convert all keys in strings (helpful if dates are keys).

    Reference: https://stackoverflow.com/questions/62198378/numeric-keys-in-yaml-files
    :param d: dictionary to convert the keys of
    :return: converted dictionary
    """
    if isinstance(d, dict):
        for idx, k in enumerate(list(d.keys())):
            if not isinstance(k, str):
                sk = str(k)
                d[sk] = d.pop(k)
                k = sk
            keys2string(d[k])
    elif isinstance(d, list):
        for e in d:
            keys2string(e)



from abc import ABC, abstractmethod
from typing import List, Union

from typeguard import typechecked

from neurokin.constants.features_extraction import VALID_EXTRACTION_TARGETS


class FeatureExtraction(ABC):

    def __new__(cls, *args, **kwargs):
        if not isinstance(cls.extraction_target, str):
            raise TypeError(f"Extraction target type not valid, expected str got {type(cls.extraction_target)}")

        if cls.extraction_target not in VALID_EXTRACTION_TARGETS:
            raise ValueError(
                f"{cls.extraction_target} in not a valid target. Valid targets are "
                f"{', '.join(str(i) for i in VALID_EXTRACTION_TARGETS)}")
        return super().__new__(cls)

    @abstractmethod
    def compute_feature(self):
        pass

    def run_feat_extraction(self, *args, **kwargs):
        feat = self.compute_feature(*args, **kwargs)
        return feat


class MyFeat(FeatureExtraction):
    extraction_target = "markers"

    @typechecked
    def __init__(self, param_a: List[Union[int, str]], param_b: List[int]):
        self.param_a = param_a
        self.param_b = param_b

    def compute_feature(self):
        print("ehi look im doing cool stuff")
        return


class MyFeat2(FeatureExtraction):
    extraction_target = "joints"

    def __init__(self, param_a: ABC, param_b: List[int]):
        self.param_a = param_a
        self.param_b = param_b

    def compute_feature(self):
        pass


if __name__ == "__main__":
    feat = MyFeat(param_a=[1, "str"], param_b=[1, 2, 3])
    feat2 = MyFeat2(param_a=3.0, param_b=[1, 2, 3])
    # feat2.run_feat_extraction()
    feat.run_feat_extraction()
    print(feat.extraction_target)
    # feat.extraction_target = "joints"
    print(feat2.extraction_target)
    # print(feat.extraction_target)



from array import array
import numpy as np

def export_neural_data_to_bin(data, filename):
    """
    Export the raw data to a binary file in C major order.

    :param data: data to be stored, should have shape n_ch*n_sample
    :param filename: name of the file
    :return:
    """
    data_flat = data.flatten(order='C')
    data_flat.tofile(filename)



import itertools
import c3d
import numpy as np
import pandas as pd


def import_c3d(path):
    """
    Fills df from c3d file

    :param path: path to .c3d
    :return: first_frame, last_frame, sample_rate, df
    """
    with open(path, "rb") as f:
        c3d_reader = c3d.Reader(f)
        first_frame = c3d_reader._header.first_frame
        last_frame = c3d_reader._header.last_frame
        sample_rate = c3d_reader._header.frame_rate
        bodyparts = get_c3d_labels(f)
        axis = ["x", "y", "z"]
        run = []
        for frame_no, points, analog in c3d_reader.read_frames(copy=False):
            fields = []
            for x, y, z, err, cam in points:
                fields.append(x)
                fields.append(y)
                fields.append(z)
            run.append(fields)
        run = np.asarray(run)
        columns = [ f"{b}_{a}" for b, a in itertools.product(bodyparts, axis)]
        df = pd.DataFrame(run, columns=columns)

    return first_frame, last_frame, sample_rate, df


#TESTME
def create_empty_df(scorer, bodyparts, frames_no):
    """
    Creates empty dataframe to receive 3d data frm c3d file

    :param scorer: mock data scorer
    :param bodyparts: list of bodyparts that will be in the dataframe
    :param frames_no: number of frames 
    :return: empty dataframe with correct shape and columns
    """
    
    dataFrame = None
    a = np.full((frames_no, 3), np.nan)
    for bodypart in bodyparts:
        pdindex = pd.MultiIndex.from_product(
            [[scorer], [bodypart], ["x", "y", "z"]],
            names=["scorer", "bodyparts", "coords"])

        frame = pd.DataFrame(a, columns=pdindex, index=range(0, frames_no))
        dataFrame = pd.concat([frame, dataFrame], axis=1)
    return dataFrame


#TESTME
def get_c3d_labels(handle):
    """
    Reads in the labels from a .c3d handle

    :param handle:
    :return: labels
    """
    reader = c3d.Reader(handle)
    a = reader._groups["POINT"]._params["LABELS"]
    C, R = a.dimensions
    labels = [a.bytes[r * C: (r + 1) * C].strip().decode().lower() for r in range(R)]
    return labels


def import_dlc_df(filename):
    if filename.endswith(".csv"):
        df = pd.read_csv(filename, header=[0, 1, 2], skipinitialspace=False, index_col=[0])
    elif filename.endswith(".hd5"):
        df = pd.read_hdf(filename)
    else:
        raise ValueError("File format not supported, please provide .csv or .hd5 files for DLC datasets")
    return df



import csv
import os
from typing import List, Tuple

import numpy as np
import pandas as pd
from pandas.errors import EmptyDataError, ParserError
from scipy import stats

from neurokin.neural_data import NeuralData
from neurokin.utils.neural import (processing, importing)


def get_csv_first_block_len(csv_path: str) -> int:
    """
    Reads in a csv and returns the len of the first block. This is until a row is empty.

    :param csv_path: path of the file
    :return: len of block
    """
    with open(csv_path, newline='') as f:
        count = 0
        rows = csv.reader(f)
        for line in rows:
            if not line:
                break
            count += 1
    return count


def get_first_block_df(csv_path: str, skiprows: int = 0) -> pd.DataFrame:
    """
    Gets the first block from a csv path. That is until the first empty line.

    :param csv_path: path of the file
    :param skiprows: how many rows to skip at the beginning
    :return: pandas dataframe of the first block
    """
    block_end = get_csv_first_block_len(csv_path) - skiprows
    df = pd.read_csv(csv_path, skiprows=skiprows, nrows=block_end, low_memory=False)

    return df


def get_first_last_frame_from_csv(csv_path: str) -> Tuple[int, int]:
    """
    Given a csv from Vicon, it will return the first and last frame of the region of interest.
    This is usually encoded as the first and last row of the second block in the file

    :param csv_path: path of the file
    :return:
    """
    with open(csv_path, newline='') as f:
        count = 0
        rows = csv.reader(f)
        for line in rows:
            if not line:
                break
            count += 1
    first_block_end = count + 5

    with open(csv_path, newline='') as f:
        count = 0
        rows = csv.reader(f)
        rowcount = 0
        for line in rows:
            if rowcount <= first_block_end:
                rowcount += 1
                continue
            elif not line:
                break
            count += 1

    df = pd.read_csv(csv_path, skiprows=first_block_end, nrows=count)
    first_frame = int(df.iloc[0, 0])
    last_frame = int(df.iloc[-1, 0])
    return first_frame, last_frame


def get_freeze_ts_bound(df: pd.DataFrame) -> Tuple[List[float], List[float]]:
    """
    Returns onset and end of a freezing event in timestamps

    :param df: events dataframe
    :return: onset and end
    """
    event_onset_idxs = df.index[(df["Name"] == "Foot Off") & (df["Context"] == "General")].tolist()
    if not len(event_onset_idxs) > 0:
        return [], []

    event_onset = [df.iloc[i]["Time (s)"] for i in event_onset_idxs]
    event_end = [df.iloc[i + 1]["Time (s)"] for i in event_onset_idxs if not i + 1 == len(df)]

    return event_onset, event_end


def get_event_timestamps_freezing_active(df: pd.DataFrame,
                                         timestamp_gait: List[List[float]]) -> List[List[float]]:
    """
    Returns freezing events after gait happened

    :param df: events dataframe
    :param timestamp_gait: list of gait timestamps
    :return: list of timestamps of freezing after gait events
    """

    event_onset, event_end = get_freeze_ts_bound(df)
    if not len(event_onset) > 0:
        return []

    events_onset_selected = []
    events_end_selected = []

    for i in range(len(event_onset)):
        if event_onset[i] >= timestamp_gait[0][0]:
            events_onset_selected.append(event_onset[i])
            events_end_selected.append(event_end[i])

    events = list(map(list, zip(events_onset_selected, events_end_selected)))

    return events


def get_event_timestamps_freezing_rest(df: pd.DataFrame,
                                       timestamp_gait: List[List[float]]) -> List[List[float]]:
    """
    Returns freezing events before gait happened

    :param df: events dataframe
    :param timestamp_gait: list of gait timestamps
    :return: list of timestamps of freezing before gait events
    """
    event_onset, event_end = get_freeze_ts_bound(df)
    if not len(event_onset) > 0:
        return []

    events_onset_selected = []
    events_end_selected = []
    for i in range(len(event_onset)):
        if len(timestamp_gait) > 0:
            if event_onset[i] > timestamp_gait[0][0]:
                break
        events_onset_selected.append(event_onset[i])
        events_end_selected.append(event_end[i])

    events = list(map(list, zip(events_onset_selected, events_end_selected)))
    return events


def get_event_timestamps_gait(df: pd.DataFrame) -> List[List[float]]:
    """
    Returns gait events timestamps

    :param df:
    :return: list of timestamps of gait
    """
    event_onset_idxs = df.index[(df["Name"] == "Foot Off") & (df["Context"].isin(["Left", "Right"]))].tolist()
    event_end_idxs = df.index[(df["Name"] == "Foot Strike") & (df["Context"].isin(["Left", "Right"]))].tolist()

    if not len(event_onset_idxs) > 0:
        return []

    if not len(event_end_idxs) > 0:
        return []

    event_end_idxs = event_end_idxs if event_end_idxs[0] > event_onset_idxs[0] else event_end_idxs[1:]

    event_onset = [df.iloc[i]["Time (s)"] for i in event_onset_idxs]
    event_end = [df.iloc[i]["Time (s)"] for i in event_end_idxs]

    events = list(map(list, zip(event_onset, event_end)))

    return events


def get_idxs_events_to_exclude(framerate: int,
                               last_frame: int,
                               ts_to_exclude_fog: List[List[float]],
                               ts_to_exclude_gait: List[List[float]]) -> List[List[int]]:
    """
    Returns the indexes corresponding to the timestamps to exclude. It merges freezing events and gait events

    :param framerate: recording framerate
    :param last_frame: last frame of the region of interest
    :param ts_to_exclude_fog: timestamps of freezing
    :param ts_to_exclude_gait: timestamps of gait
    :return: indexes of frames to be excluded
    """
    idxs_to_exclude = ts_to_exclude_fog + ts_to_exclude_gait
    for idx_pair in idxs_to_exclude:
        if np.isnan(idx_pair[1]):
            idx_pair[1] = last_frame
    idxs_to_exclude_frames = [[int(i[0] * framerate), int(i[1] * framerate)] for i in idxs_to_exclude]
    idxs_to_exclude_frames = sorted(idxs_to_exclude_frames, key=lambda x: x[0])

    return idxs_to_exclude_frames


def get_start_of_gait(framerate: int, ts_gait: List[List[float]]) -> int:
    """
    Returns the first frame of gait

    :param framerate: recording framerate
    :param ts_gait: timestamps of gait
    :return: first frame index of gait
    """
    idxs_gait = [[int(i[0] * framerate), int(i[1] * framerate)] for i in ts_gait]
    start_gait = idxs_gait[0][0]
    return start_gait


def traspose_idxs(idxs_to_exclude: List[List[int]], first_frame: int) -> List[List[int]]:
    """
    Transposes the indexes to an array where first_frame is 0. Moreover it takes care of small idiosyncrasies when
    an event appears to start at frame -1

    :param idxs_to_exclude: list of indexes to be set to false
    :param first_frame: first frame of the region of interest
    :return: indexes transposed, same shape as inut
    """
    idxs_transposed = []

    for idx_pair in idxs_to_exclude:
        start = idx_pair[0] - first_frame if idx_pair[0] - first_frame >= 0 else 0
        end = idx_pair[1] - first_frame
        idxs_transposed.append([start, end])
    return idxs_transposed


def transpose_start_of_gait(start_of_gait: int, first_frame: int) -> int:
    """
    Transposes the index of the start of gait where first_frame is 0.
    Moreover it takes care of small idiosyncrasies when an event appears to start at frame -1

    :param start_of_gait: start of the first gait event
    :param first_frame: first frame of the region of interest
    :return:
    """
    return start_of_gait - first_frame if start_of_gait - first_frame > 0 else 0


def create_exclusion_mask(idxs_to_exclude: List[List[int]], first_frame: int, last_frame: int) -> np.array:
    """
    Returns an array representing the region of interest of a run with
    True values only where there are no events to exclude

    :param idxs_to_exclude: list of the pairs of indexes to exclude [start, end]
    :param first_frame: first frame of the region of interest
    :param last_frame: last frame of the region of interest
    :return: exclusion mask
    """
    idxs_to_exclude = traspose_idxs(idxs_to_exclude, first_frame)
    full_idx_array = np.arange(first_frame, last_frame)
    mask = np.ones(full_idx_array.size, dtype=bool)

    # ensuring that there is a bool change even if the run starts or ends with a nlm
    mask[0] = False
    mask[-1] = False

    for idxs_ex in idxs_to_exclude:
        mask[idxs_ex[0]:idxs_ex[1]] = False

    return mask


def get_ts_from_exclusion_mask(mask: np.array, first_frame: int, framerate: int) -> Tuple[List[float], List[float]]:
    """
    Gets bounds of the exclusion mask and converts it in timestamps

    :param mask:
    :param first_frame: first frame of the region of interest
    :param framerate: recording framerate
    :return:
    """
    events_bounds = np.where(np.diff(mask))[0]

    event_onset_idxs = events_bounds[0::2] + first_frame
    event_end_idxs = events_bounds[1::2] + first_frame

    event_onset = [i / framerate for i in event_onset_idxs]
    event_end = [i / framerate for i in event_end_idxs]
    return event_onset, event_end


def get_event_timestamps_nlm_active(framerate: int,
                                    first_frame: int,
                                    last_frame: int,
                                    ts_to_exclude_fog: List[List[float]],
                                    ts_to_exclude_gait: List[List[float]]):
    """
    Creates a mask representing the region of interest excluding all the passed events,
    and any event before the first gait frame

    :param framerate: recording framerate
    :param first_frame: first frame of the region of interest
    :param last_frame: last frame of the region of interest
    :param ts_to_exclude_fog: timestamps of freezing
    :param ts_to_exclude_gait: timestamps of gait
    :return: non-locomotion movements event after gait has happened
    """
    idxs_to_exclude = get_idxs_events_to_exclude(framerate,
                                                 last_frame,
                                                 ts_to_exclude_fog,
                                                 ts_to_exclude_gait)

    if first_frame is None:
        first_frame = idxs_to_exclude[0][0]
        last_frame = idxs_to_exclude[-1][1]

    mask = create_exclusion_mask(idxs_to_exclude, first_frame, last_frame)

    if len(ts_to_exclude_gait) > 0:
        start_gait = get_start_of_gait(framerate, ts_to_exclude_gait)
        start_gait = transpose_start_of_gait(start_gait, first_frame)
        mask[:start_gait] = False  # setting to False all frames before gait

    event_onset, event_end = get_ts_from_exclusion_mask(mask, first_frame, framerate)

    events = list(map(list, zip(event_onset, event_end)))
    return events


def get_event_timestamps_nlm_rest(framerate: int,
                                  first_frame: int,
                                  last_frame: int,
                                  ts_to_exclude_fog: List[List[float]],
                                  ts_to_exclude_gait: List[List[float]]):
    """
    Creates a mask representing the region of interest excluding all the passed events,
    and any event after the first gait frame

    :param framerate: recording framerate
    :param first_frame: first frame of the region of interest
    :param last_frame: last frame of the region of interest
    :param ts_to_exclude_fog: timestamps of freezing
    :param ts_to_exclude_gait: timestamps of gait
    :return: non-locomotion movements event after gait has happened
    """
    idxs_to_exclude = get_idxs_events_to_exclude(framerate,
                                                 last_frame,
                                                 ts_to_exclude_fog,
                                                 ts_to_exclude_gait)

    if first_frame is None:
        first_frame = idxs_to_exclude[0][0]
        last_frame = idxs_to_exclude[-1][1]

    mask = create_exclusion_mask(idxs_to_exclude, first_frame, last_frame)

    if len(ts_to_exclude_gait) > 0:
        start_gait = get_start_of_gait(framerate, ts_to_exclude_gait)
        start_gait = transpose_start_of_gait(start_gait, first_frame)
        mask[start_gait:] = False  # setting to False all frames after gait

    event_onset, event_end = get_ts_from_exclusion_mask(mask, first_frame, framerate)

    events = list(map(list, zip(event_onset, event_end)))
    return events


def get_neural_correlate_psd(raw: np.array,
                             fs: float,
                             t_on: float,
                             t_off: float,
                             nfft: int,
                             nov: int,
                             zscore: bool = True) -> Tuple[np.array, np.array]:
    """
    Takes the beginning and end of an event, retrieves the corresponding chunk of neural data and computes the psd

    :param raw: raw neural single channel
    :param fs: sampling frequency
    :param t_on: beginning of event in seconds
    :param t_off: end of event in seconds
    :param nfft: nfft for fourier transform
    :param nov: overlap for fourier transform
    :param zscore: how to normalize
    :return: psd and corresponding frequencies
    """
    s_on = importing.time_to_sample(t_on, fs=fs, is_t1=True)
    s_off = importing.time_to_sample(t_off, fs=fs, is_t2=True)

    neural_event = raw[s_on:s_off]

    freqs, pxx = processing.calculate_power_spectral_density(neural_event, fs, nperseg=nfft, noverlap=nov,
                                                             scaling="spectrum")

    if zscore:
        pxx = stats.zscore(pxx)

    return pxx, freqs


def get_neural_ch(neural_path: str, ch: int, stream_name: str) -> Tuple[np.array, float]:
    """
    Retireves the raw neural signal and the sampling frequency

    :param neural_path: folder path to TDT recording
    :param ch: channel number 0 based
    :param stream_name: name  of the Stored listing
    :return: raw neural single channel and sampling frequency
    """
    neural = NeuralData(path=neural_path)
    neural.load_tdt_data(stream_name=stream_name)

    return neural.raw[ch], neural.fs


def time_to_frame_in_roi(timestamp: float, fs: float, first_frame: int) -> int:
    """
    Converts the time to frames in the region of interest (i.e. with an offset of the first frame)

    :param timestamp: time in seconds
    :param fs: sampling frequency
    :param first_frame: first frame of the region of interest
    :return:
    """
    frame = int(timestamp * fs) - first_frame
    frame = 0 if frame == -1 else frame
    if frame < 0:
        raise ValueError(f"First frame value is bigger than frame of event. "
                         f"First frame is {first_frame} and event frame is {frame}")

    return frame


def get_events_dict(event_path, skiprows, framerate):
    """
    Runs trough all the .csv files in the experiment_path and returns a dictionary with the timestamps of each
    event, sorted by the type.

    :param event_path: path to the run to be analyzed
    :param skiprows: how many rows to skip as a header of the .csv
    :param framerate: frame rate of the acquisition system
    :return: events dictionary
    """
    events_dict = {}
    events_dict.setdefault("gait", [])
    events_dict.setdefault("fog_active", [])
    events_dict.setdefault("fog_rest", [])
    events_dict.setdefault("nlm_active", [])
    events_dict.setdefault("nlm_rest", [])

    df = get_first_block_df(csv_path=event_path, skiprows=skiprows)

    try:
        first_frame, last_frame = get_first_last_frame_from_csv(event_path)
    except EmptyDataError:
        first_frame, last_frame = None, None
    except ParserError:
        first_frame, last_frame = None, None
    try:
        df.sort_values('Time (s)', inplace=True, ignore_index=True)
    except KeyError:
        print("No events were labelled, assuming non-locomotor movement on the whole run")
        event_onset, event_end = [first_frame / framerate], [last_frame / framerate]
        events_dict["nlm_rest"] = list(map(list, zip(event_onset, event_end)))
        return events_dict

    events_dict["gait"] = get_event_timestamps_gait(df)

    if len(events_dict["gait"]) > 0:
        events_dict["fog_active"] = get_event_timestamps_freezing_active(df, events_dict["gait"])

    events_dict["fog_rest"] = get_event_timestamps_freezing_rest(df, events_dict["gait"])

    events_detected_fog = events_dict["fog_active"] + events_dict["fog_rest"]

    if len(events_dict["gait"]) > 0:
        events_dict["nlm_active"] = get_event_timestamps_nlm_active(framerate=framerate,
                                                                    first_frame=first_frame,
                                                                    last_frame=last_frame,
                                                                    ts_to_exclude_fog=events_detected_fog,
                                                                    ts_to_exclude_gait=events_dict["gait"])

    events_dict["nlm_rest"] = get_event_timestamps_nlm_rest(framerate=framerate,
                                                            first_frame=first_frame,
                                                            last_frame=last_frame,
                                                            ts_to_exclude_fog=events_detected_fog,
                                                            ts_to_exclude_gait=events_dict["gait"])

    return events_dict


def get_neural_correlates_dict(neural_path,
                               channel_of_interest,
                               stream_names,
                               events_df,
                               time_cutoff):
    """
    Returns a dictionary with the raw neural data corresponding to each event in the input events dictionary, from the
    selected channel of interest. Events duration is capped at time_cutoff to ensure homogeneity in the duration.

    :param neural_path: path to the folder with neural data
    :param channel_of_interest: channel to extract the raw data from
    :param stream_names: list of possible stream_names where the neural data is stored
    :param events_df:
    :param time_cutoff:
    :return:
    """
    states = [key for key in events_df.columns if key.startswith("event")]
    neural_dict = {key: [] for key in states}
    fs = None
    print(neural_path)
    try:
        neural_path = neural_path + next(os.walk(neural_path))[1][0]
    except (StopIteration, IndexError) as e:
        print(f"{e} No neural data found for {neural_path}")
        return neural_dict, fs

    is_valid_name = False
    for name in stream_names:
        try:
            raw, fs = get_neural_ch(neural_path, channel_of_interest, name)
            is_valid_name = True
            break
        except (AttributeError, TypeError):
            pass
    if not is_valid_name:
        raise Exception("All stream names were invalid")

    for state in states:
        neural_dict[state] = get_single_neural_type(events_df, state, time_cutoff, fs, raw)

    return neural_dict, fs


def get_single_neural_type(events_df, event_type, time_cutoff, fs, raw):
    """
    Fetches the neural chunks corresponding to all the events of a certain type. All events
    are cut at the time_cutoff, to ensure consistency across different lengths (this implies
    events shorter than time_cutoff are dropped).

    :param events_df: dataframe containing all events timestamps
    :param event_type: type of event to fetch the neural correlates for
    :param time_cutoff: max length of an event to consider (shorter events are dropped)
    :param fs: sampling frequency of the neural data
    :param raw: raw neural data, single channel
    :return: returns a list of the neural chunks corresponding to the events, cropped to time_cutoff
    """
    correlates = []
    states_events_list = events_df[event_type].values
    for states_events in states_events_list:
        for t_onset, t_end in states_events:
            t_end = check_time_cutoff(t_onset, t_end, time_cutoff)
            if t_end is None:
                continue
            else:
                s_on = importing.time_to_sample(t_onset, fs=fs, is_t1=True)
                s_off = importing.time_to_sample(t_end, fs=fs, is_t2=True)
                correlates.append(raw[s_on:s_off])
                if len(raw[s_on:s_off]) == 0:
                    raise ValueError("Warning the neural chunk is of 0 length, "
                                     "check match between neural and states files")

    return correlates


def compute_psd_for_row(row, events_columns, nfft, noverlap, zscore):
    """
    Computes Power Spectra Density arrays for all the events columns.

    :param row: a pd.Dataframe row, containing a list of neural correlates for each event type and fs
    :param events_columns: which columns contain the lists of neural chunks
    :param nfft: what number of fft segments to use
    :param noverlap: the overlap of the fft segments
    :param zscore: whether to z-score the psd or not
    :return: Returns a pd.Series containing lists of Power Spectra Density arrays for all the events columns.
    """
    fs = row['fs']
    psd_results = {}
    for event in events_columns:
        psd_results[event] = get_psd_single_event_type(row[event], fs=fs, nfft=nfft, noverlap=noverlap, zscore=zscore)
    return pd.Series(psd_results)


def get_psd_single_event_type(raw_neural_list, fs, nfft, noverlap, zscore):
    """
    Computes the Power Spectra Density array for all element in the list (corresponding to a single type of event)

    :param raw_neural_list: list containing neural data arrays
    :param fs: sampling frequency of the neural data
    :param nfft: what number of fft segments to use
    :param noverlap: the overlap of the fft segments
    :param zscore: whether to z-score the psd or not
    :return: list of PSDs
    """
    psds = []
    freqs = None
    for raw_neural in raw_neural_list:
        freqs_psd, pxx = processing.calculate_power_spectral_density(raw_neural,
                                                                     fs,
                                                                     nperseg=nfft,
                                                                     noverlap=noverlap,
                                                                     scaling="spectrum")
        if freqs is None:
            freqs = freqs_psd

        sanity_check = np.array_equal(freqs, freqs_psd)
        if not sanity_check:
            raise ValueError("The frequencies in the PSD calculation are unequal for different events. "
                             "Check for consistency in the events length")
        if zscore:
            pxx = stats.zscore(pxx)
        psds.append(pxx)
    return psds


def check_time_cutoff(t_onset, t_end, time_cutoff):
    """
    Checks if the difference between t_onset and t_end is lower than the time_cutoff.
    If not then computes the end edge as t_onset + t_end and returns it.

    :param t_onset: start of the event
    :param t_end: end of the event
    :param time_cutoff: minimum length of the event
    :return: None if event is too short, t_onset + t_end otherwise
    """
    if t_end - t_onset < time_cutoff:
        return None
    elif t_end - t_onset > time_cutoff:
        t_end = t_onset + time_cutoff
        return t_end



from importlib import import_module
from typing import List
import pandas as pd

from neurokin.constants.features_extraction import FEATURES_EXTRACTION_MODULE


def get_extractor_obj(feature_name):
    module_, feature_class = feature_name.rsplit(".", maxsplit=1)
    module_ = FEATURES_EXTRACTION_MODULE + module_
    m = import_module(module_)
    feature_extract_class = getattr(m, feature_class)
    return feature_extract_class()


def extract_features(features, bodyparts, skeleton, markers_df, get_binned, bin_params, custom_feats):
    extracted_features = []
    binned_features = []
    binned_df = None
    for feature_name, params in features.items():
        params = params if params else {}
        params["custom_features"] = custom_feats
        extractor_obj = get_extractor_obj(feature_name)
        extraction_target = extractor_obj.extraction_target

        if extraction_target == "markers":
            target_bodyparts = params.get("marker_ids", bodyparts)

        elif extraction_target == "joints":
            target_joints = params.get("joint_ids", skeleton[extraction_target].keys())
            target_bodyparts = {joint: skeleton[extraction_target][joint] for joint in target_joints}

        elif extraction_target == "misc":
            target_bodyparts = params.get("misc_ids", bodyparts)

        else:
            raise ValueError(f"{extraction_target} is not a valid extraction target."
                             f"Please use: markers, joints or multiple_markers")

        feature = extractor_obj.run_feat_extraction(df=markers_df, target_bodyparts=target_bodyparts, **params)
        extracted_features.append(pd.DataFrame(feature).reset_index(drop=True))
        if get_binned:
            try:
                binning_strategy = params["binning_strategy"]
            except KeyError:
                continue
            binned_features.append(bin_feature(feature,
                                               binning_strategies=binning_strategy,
                                               window=bin_params["window_size"],
                                               overlap=bin_params["overlap"]).reset_index(drop=True))
            binned_df = pd.concat(binned_features, axis=1)
    feats_df = pd.concat(extracted_features, axis=1)
    return feats_df, binned_df


def bin_feature(feature, binning_strategies: List[str], window, overlap):
    step = window - overlap
    binned_features = []
    if step <= 0:
        raise ValueError(f"The overlap should be lower than the window. Got overlap: {overlap} and window: {window}.")
    for strategy in binning_strategies:
        if strategy.lower().strip(" ") == "mean":
            binned = feature.rolling(window=window, step=step).mean().add_suffix("_mean")
        elif strategy.lower().strip(" ") == "median":
            binned = feature.rolling(window=window, step=step).median().add_suffix("_median")
        elif strategy.lower().strip(" ") == "min":
            binned = feature.rolling(window=window, step=step).min().add_suffix("_min")
        elif strategy.lower().strip(" ") == "max":
            binned = feature.rolling(window=window, step=step).max().add_suffix("_max")
        elif strategy.lower().strip(" ") == "sum":
            binned = feature.rolling(window=window, step=step).sum().add_suffix("_sum")
        elif strategy.lower().strip(" ") == "std":
            binned = feature.rolling(window=window, step=step).std().add_suffix("_std")
        else:
            raise ValueError(f"The chose binning strategy: {strategy} is not available. Please choose between:"
                             f"mean, median, min, max, sum or std")
        binned_features.append(binned)
    binned_features_df = pd.concat(binned_features, axis=1)
    return binned_features_df



import warnings
import numpy as np
from numpy.typing import ArrayLike
from scipy import signal
from typing import List, Tuple, Union
from neurokin.utils.neural.importing import time_to_sample


def simply_mean_data_binarize(sync_ch: np.ndarray):
    """
    Return a simply binarized array by setting anything below the mean to 0 and everything above to 1.
    Use only in very clear-cut cases

    :param sync_ch: input sync channel
    :return: binarized array
    """
    if np.any(np.isnan(sync_ch)):
        raise ValueError("The input array contains nan, please cast them to a number using np.nan_to_num or equivalent")
    mean = np.mean(sync_ch)
    binarized = np.where(sync_ch < mean, 0, 1)
    return binarized


def get_stim_timestamps(sync_ch: np.ndarray, expected_pulses: int = None) -> np.ndarray:
    """
    Get indexes of only threshold crossing up from 0, i.e. edge detection.
    Sometimes there are spurious signals on the stimulation channel, when the stimulator turns off;
    if expected_pulses is given this function trims to the expected number of pulses.

    :param sync_ch: the stimulation sync channel data
    :param expected_pulses: number of expected pulses
    :return: trimmed indexes
    """

    def process_channel(sync_ch):
        if np.any(np.isnan(sync_ch)):
            raise ValueError(
                "The input array contains nan, please cast them to a number using np.nan_to_num or equivalent")
        threshold_crossing = np.diff(sync_ch > 0, prepend=False)
        idxs_edges = np.where(threshold_crossing)[0]
        stim_starts = idxs_edges[::2]
        if expected_pulses is not None:
            if not isinstance(expected_pulses, int) or expected_pulses <= 0:
                raise ValueError("The expected pulses must be a positive integer")
            if expected_pulses < len(stim_starts):
                warnings.warn(
                    "Warning: number of pulses is greater than the expected pulses, trimming to the expected pulses. "
                    "(expected behaviour by function design)")
            if expected_pulses > len(stim_starts):
                warnings.warn(
                    "Warning: number of pulses is less than the expected pulses, returning all the pulses found. "
                    "(expected behaviour by function design)")
            return stim_starts[:expected_pulses]
        else:
            return stim_starts

    if sync_ch.ndim == 1:
        return process_channel(sync_ch)
    elif sync_ch.ndim == 2:
        return np.array([process_channel(row) for row in sync_ch], dtype=object)
    else:
        raise ValueError("sync_ch must be a 1D or 2D array")


def get_timestamps_stim_blocks(neudata, n_amp_tested, pulses, time_stim):
    """
    Given a DBS recording with multiple stimulation amplitudes tested it gives the time stamps of the onset and end
    of each block of stimulation.

    :param neudata: NeuralData object, containing sync_data
    :param n_amp_tested:
    :param pulses:
    :param time_stim:
    :return:
    """
    if len(neudata.sync_data.shape) > 1:
        raise ValueError("Warning: depending on the experiment setting the sync data might be a multichannel array \n"
                         "please use the .pick_sync_data(idx) method to pick the correct sync channel.")
    total_pulses = n_amp_tested * pulses
    sync = get_stim_timestamps(neudata.sync_data, total_pulses)
    onset = sync[0:len(sync):pulses]
    stim_len = time_to_sample(time_stim, neudata.fs, is_t2=True)
    # FIXME check that neudata.fs is correct if stim and raw have different fs
    timestamps_blocks = [(onset[i], onset[i] + stim_len) for i in range(len(onset))]
    return timestamps_blocks


def get_median_distance(a: ArrayLike) -> float:
    """
    Gets median distance between points

    :param a: array-like data
    :return: median
    """
    if not isinstance(a, np.ndarray):
        raise TypeError("Input must be an array like type")
    distances = abs(np.diff(a))
    return np.median(distances)


def running_mean(x: ArrayLike, n: int) -> ArrayLike:
    """
    Returns the running mean with window n

    :param x: data
    :param n: window size
    :return: smoothed data
    """
    x = np.asarray(x)
    if x.ndim != 1:
        raise ValueError("Input must be a 1D array")
    if n <= 0:
        raise ValueError("Window size must be a positive integer")
    if n > x.shape[0]:
        raise ValueError("Window size must be smaller than the data length")
    cumsum = np.cumsum(np.insert(x, 0, 0))
    return (cumsum[n:] - cumsum[:-n]) / float(n)

def running_mean_2D(x: ArrayLike, n: int) -> ArrayLike:
    """
        Returns the running mean with window n, on 2D data

        :param x: data
        :param n: window size
        :return: smoothed data
        """
    smoothed_data = np.apply_along_axis(running_mean, 1, x, n)
    return smoothed_data


def trim_equal_len(raw: List[ArrayLike]) -> List[float]:
    """
    Trims a list of arrays to all have the same length.

    :param raw: raw data
    :return: list with trimmed data
    """
    if not all(np.asarray(arr).ndim == 1 for arr in raw):
        raise ValueError("Not all arrays are 1D")
    lens = [len(r) for r in raw]
    equaled = [r[:min(lens)] for r in raw]
    return equaled


def parse_raw(raw: np.ndarray, stimulation_idxs: np.ndarray, samples_before_stim: int,
              skip_one: bool = False, min_len_chunk: int = 1) -> np.ndarray:
    """
    Parses the signal given the timestamps of the stimulation.

    :param raw: raw data of one channel
    :param stimulation_idxs: indexes of the stimulation onset
    :param samples_before_stim: how much before the stimulation onset to parse
    :param skip_one: if True parses every second stimulation
    :param min_len_chunk: filters the chunks to have a minimal len between pulses
    :return: parsed raw signal into an array of equally sized chunks
    """
    stimulation_idxs = np.asarray(stimulation_idxs)
    if len(stimulation_idxs) == 0:
        raise ValueError("Stimulation indexes is empty, please check the input data")
    stimulation_idxs = stimulation_idxs - samples_before_stim
    if stimulation_idxs[0] < 0:
        stimulation_idxs[0] = 0

    if skip_one:
        stimulation_idxs = stimulation_idxs[::2]
    split_raw = np.split(raw, stimulation_idxs)[
                1:]  # skip chunk that precedes the first stimulation to avoid cropping errors
    split_raw = [chunk for chunk in split_raw if len(chunk) >= min_len_chunk]
    leveled_list = trim_equal_len(split_raw)
    trimmed_array = np.vstack(leveled_list)
    return trimmed_array


# TESTME
def get_average_amplitudes(parsed_raw, tested_amplitudes, pulses_number=None):
    """
    Assumes an experiment where a sequence of stimulation amplitudes are applied,
    every stimulation with a fixed number of pulses

    :param parsed_raw:
    :param tested_amplitudes:
    :param pulses_number:
    :return:
    """
    number_tested_amplitudes = len(tested_amplitudes)
    averaged_amplitudes = []
    if number_tested_amplitudes == 1:
        averaged_amp = average_block(parsed_raw, 0, len(parsed_raw))
        return [averaged_amp]
    if not pulses_number:
        pulses_number = int(len(parsed_raw) / len(tested_amplitudes))
    for i in range(number_tested_amplitudes):
        start = i * pulses_number
        stop = (i + 1) * pulses_number
        averaged_amp = average_block(parsed_raw, start, stop)
        averaged_amplitudes.append(averaged_amp)
    return averaged_amplitudes


# TESTME
def average_block(array: ArrayLike, start: int, stop: int) -> np.ndarray:
    """
    Averages a subset of elements

    :param array: input array
    :param start: start idx for parsing
    :param stop: stop idx for parsing
    :return: average of the subset
    """
    parsed = array[start:stop]
    averaged = np.mean(parsed, axis=0)
    return averaged


def find_closest_index(data: ArrayLike, datapoint: float) -> int:
    """
    Given an array of data and a datapoint it returns the index of the element that has
    the minimum difference to the datapoint

    :param data: data array-like
    :param datapoint: datapoint to find a close value to
    :return: index of the closest element
    """
    data = np.asarray(data)
    if np.any(np.isnan(data)):
        raise ValueError('The input array contains nan which will always return as the nearest to datapoint'
                         'please cast it to a number using np.nan_to_num or equivalent')
    diff = np.abs(data - datapoint)

    if data.ndim == 1:
        return diff.argmin()
    elif data.ndim == 2:
        flat_idx = diff.argmin()
        return np.unravel_index(flat_idx, data.shape)
    else:
        raise ValueError("Input must be a 1D or 2D array")


# TESTME

def find_closest_smaller_index(data: ArrayLike, datapoint: float) -> Union[int, Tuple[int, int]]:
    """
    Given an array of data and a datapoint, returns the index of the element that is
    the closest but lower than the datapoint.

    For a 1D array, returns an integer index.
    For a 2D array, returns a tuple (row, column) indicating the location of the closest element.

    :param data: data array-like (1D or 2D)
    :param datapoint: datapoint to find a close value to
    :return: index of the closest element below datapoint (int for 1D or tuple for 2D)
    """
    data = np.asarray(data)
    if np.any(np.isnan(data)):
        raise ValueError("The input array contains nan; please cast them to a number using np.nan_to_num or equivalent")

    mask = data < datapoint
    differences = np.where(mask, datapoint - data, np.inf)

    if np.all(np.isinf(differences)):
        raise ValueError("No element in the array is smaller than the datapoint")

    flat_idx = differences.argmin()

    if data.ndim == 1:
        return flat_idx
    elif data.ndim == 2:
        return np.unravel_index(flat_idx, data.shape)
    else:
        raise ValueError("Input must be a 1D or 2D array")


# TESTME with TDT data
def get_spectrogram_data(fs: float, raw: ArrayLike, nfft: int = None,
                         noverlap: int = None) -> Tuple[ArrayLike]:
    """
    Gets the data used to plot a spectrogram

    :param fs: sampling frequency
    :param raw: raw data
    :param nfft: nfft to compute the fft
    :param noverlap: number of overlap points
    :return:
    """
    _t, _f, sxx = signal.spectrogram(raw, fs=fs, nperseg=nfft, noverlap=noverlap)

    return sxx, _f, _t


# TESTME with TDT data
def calculate_power_spectral_density(data: ArrayLike, fs: float, **kwargs) -> tuple:
    """
    Calculate the frequencies and power spectral densities from the raw recording time series data.

    :param data: raw recording data
    :param fs: sampling frequency
    :return: frequencies, power spectral density
    """
    freq, pxx = signal.welch(data, fs, **kwargs)
    return freq, pxx





from typing import List

import numpy as np
import pandas as pd
from typeguard import typechecked

from neurokin.utils.features_extraction.core import FeatureExtraction
from neurokin.utils.features_extraction import commons


class LinearVelocity(FeatureExtraction):
    extraction_target = "markers"

    @typechecked
    def compute_feature(self, df: pd.DataFrame, target_bodyparts: List, **kwargs):
        bodyparts_coordinates = df.columns.tolist()
        target_markers_coords = [coord for marker in target_bodyparts for coord in bodyparts_coordinates if
                                 marker in coord]
        column_names = {bodypart: f"{bodypart}_linear_velocity" for bodypart in target_markers_coords}

        df_feat = df[target_markers_coords].apply(commons.compute_velocity)
        df_feat.rename(columns=column_names, inplace=True)

        return df_feat


class LinearAcceleration(FeatureExtraction):
    extraction_target = "markers"

    @typechecked
    def compute_feature(self, df: pd.DataFrame, target_bodyparts: List, **kwargs):
        bodyparts_coordinates = df.columns.tolist()
        target_markers_coords = [coord for marker in target_bodyparts for coord in bodyparts_coordinates if
                                 marker in coord]
        column_names = {bodypart: f"{bodypart}_linear_acceleration" for bodypart in target_markers_coords}

        df_feat = df[target_markers_coords].apply(commons.compute_acceleration)
        df_feat.rename(columns=column_names, inplace=True)

        return df_feat


class LinearSpeed(FeatureExtraction):
    extraction_target = "markers"

    @typechecked
    def compute_feature(self, df: pd.DataFrame, target_bodyparts: List, **kwargs):
        bodyparts_coordinates = df.columns.tolist()
        feature_df_list = []
        df_feat = pd.DataFrame()
        for marker in target_bodyparts:
            coords = [coord for coord in bodyparts_coordinates if marker in coord]
            feat = commons.compute_speed(df[coords])
            df_feat = pd.DataFrame(feat, columns=[f"{marker}_linear_speed"])
            feature_df_list.append(df_feat)

        if feature_df_list:
            df_feat = pd.concat(feature_df_list, axis=1)
        return df_feat


class TangentialAcceleration(FeatureExtraction):
    extraction_target = "markers"

    @typechecked
    def compute_feature(self, df: pd.DataFrame, target_bodyparts: List, **kwargs):
        bodyparts_coordinates = df.columns.tolist()
        feature_df_list = []
        df_feat = pd.DataFrame()
        for marker in target_bodyparts:
            coords = [coord for coord in bodyparts_coordinates if marker in coord]
            feat = commons.compute_tang_acceleration(df[coords])
            df_feat = pd.DataFrame(feat, columns=[f"{marker}_tang_acceleration"])
            feature_df_list.append(df_feat)

        if feature_df_list:
            df_feat = pd.concat(feature_df_list, axis=1)
        return df_feat



import numpy as np
import numpy as np
import pandas as pd
from numpy.typing import NDArray


def compute_angle(vectors):
    """
    Computes the angle between 3 points in a 3d or 2d space.
    :param vectors: input coordinates
    """
    vectors = np.asarray(vectors)
    if vectors.ndim == 1:
        if vectors.shape[0] in (6, 9):
            vectors = vectors.reshape(1, -1)
        else:
            raise IndexError(f"Expected 1D array of length 6 or 9, but got length {vectors.shape[0]}.")

    try:
        if vectors.shape[1] == 9:
            a, b, c = vectors[:, :3], vectors[:, 3:6], vectors[:, 6:9]
        elif vectors.shape[1] == 6:
            a, b, c = vectors[:, :2], vectors[:, 2:4], vectors[:, 4:6]
    except IndexError:
        raise IndexError("The angles can only be computed on 3 points in a 3d or 2d space. The vectors shape should"
                         f"be (n, 6) or (n, 9) but got shape {vectors.shape}")

    bas = b - a
    bcs = c - b
    angles = []
    for ba, bc in zip(bas, bcs):
        cosine_angle = np.dot(ba, bc) / (np.linalg.norm(ba) * np.linalg.norm(bc))
        a = np.arccos(cosine_angle)
        angles.append(a)
    return np.array(angles)


def compute_angle_correlation(vectors):
    """
    Computes the correlation between the angles of a set of vectors.
    :param vectors: input coordinates
    """
    angles = compute_angle(vectors)
    angle_corr = np.corrcoef(angles)
    return angle_corr


def compute_angle_velocity(vectors):
    """
    Computes the velocity of the angle of a set of vectors.
    :param vectors: input coordinates
    """
    angles = compute_angle(vectors)
    angle_vel = np.gradient(angles, 1)
    return angle_vel


def compute_angle_acceleration(vectors):
    """
    Computes the acceleration of the angle of a set of vectors.
    :param vectors: input coordinates
    """
    angle_acc = np.gradient(compute_angle_velocity(vectors), 1)
    return angle_acc


def compute_angle_phase(vectors):
    """
    Computes the phase of a signal.
    First computes the fft, then the phase from the complex element at each index.
    :param vectors: input coordinates
    :return: phases
    """
    angles = compute_angle(vectors)
    spectrum = np.fft.fft(angles)
    phase = np.angle(spectrum)
    return phase


def compute_phase_at_max_amplitude(vectors):
    """
    Computes the phase of a signal at frequency that has the maximum amplitude.
    First computes the fft, then gets the index of maximum value from the real component, then the phase from the
    complex element at that index.

    :param vectors: input coordinates
    :return: phase
    """
    angles = compute_angle(vectors)
    c_transform = np.fft.fft(angles)
    r_transform = abs(c_transform) ** 2
    freq_of_interest = np.argmax(r_transform)
    phase = np.angle(c_transform[freq_of_interest], deg=True)
    phase = phase if phase > 0 else 360 + phase
    return phase


def compute_speed(df: pd.DataFrame) -> NDArray:
    """
    Computes the speed of a trajectory.
    :param df: input coordinates
    """
    traj = df.apply(compute_velocity)
    speed = np.apply_along_axis(np.linalg.norm, 1, traj, None, 0)
    return speed


def compute_velocity(df: pd.DataFrame) -> NDArray:
    """
    Computes the velocity of a trajectory.
    :param df: input coordinates
    """
    df = df.values
    return np.gradient(df, 1)


def compute_acceleration(df: pd.DataFrame) -> NDArray:
    """
    Computes the acceleration of a trajectory.
    :param df: input coordinates
    """
    velocity = compute_velocity(df)
    return np.gradient(velocity, 1)


def compute_tang_acceleration(df: pd.DataFrame) -> NDArray:
    """
    Computes the tangential acceleration of a trajectory.
    :param df: input coordinates
    """
    speed = compute_speed(df)
    return np.gradient(speed, 1)
